/*
 * Copyright 2011-2017 https://pvoutput.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.pvoutput.integration.reader;

import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.TimeZone;

import org.apache.log4j.Logger;
import org.json.simple.JSONObject;
import org.json.simple.JSONValue;
import org.pvoutput.integration.Output;
import org.pvoutput.integration.Sensor;
import org.pvoutput.integration.util.WebClient;


public class PowerwallLogReader extends AHttpLogReader
{	
	private static Logger LOGGER = Logger.getLogger(PowerwallLogReader.class);

	private SimpleDateFormat df = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
		
	public void startup()
	throws Exception
	{
		super.startup();
	
		PowerwallFileWriter w = new PowerwallFileWriter();
		w.start();
		
		df.setTimeZone(TimeZone.getTimeZone("UTC"));
	}
	
	public String getName()
	{
		return "powerwall";
	}
	
	public AHttpLogReader copy()
	{
		return new PowerwallLogReader();
	}
	
	public String getLogFilePrefix()
	{
		return "PW";
	}
	
	@Override
	public int getDefaultDirection() 
	{
		return IMPORT;
	}
	
	private class PowerwallFileWriter extends SimpleLogFileWriter
	{
		public List<Output> parse(Sensor sensor, String s)
		{
			List<Output> outputs = new ArrayList<Output>();

			if(LOGGER.isDebugEnabled())
			{
				LOGGER.info("<<< " + s);
			}
			
			try
			{
				JSONObject jsonObject = (JSONObject)JSONValue.parse(s);
				
				if(jsonObject != null)
				{
					// generation
					if(sensor.direction == EXPORT)
					{
						if(jsonObject.containsKey("solar"))
						{
							JSONObject o = (JSONObject)jsonObject.get("solar");
							Date timestamp = null;
							
							Object otime = o.get("last_communication_time");
							
							if(otime instanceof String)
							{
								String stime = (String)otime;
								
								if(stime.length() >= 19)
								{
									try
									{
										if(!stime.endsWith("Z"))
										{
											// non-utc timezone
											int toffset = stime.lastIndexOf('+');
											
											String tz = "";
											
											if(toffset > 0)
											{
												tz = stime.substring(toffset, stime.length());
											}
											else
											{
												toffset = stime.lastIndexOf('-');
												
												if(toffset > 0)
												{
													tz = stime.substring(toffset, stime.length());
												}
											}
											
											df.setTimeZone(TimeZone.getTimeZone("GMT" + tz));
										}
										
										timestamp = df.parse(stime.substring(0, 19));
									}
									catch(Exception e)
									{
										LOGGER.error("Invalid date format: " + stime.substring(0, 19));
									}
								}
							}
							
							if(timestamp != null)
							{
								String key = (String)properties.get("power");
								
								if(key == null)
								{
									key = "instant_power";
								}
								
								Object p = o.get(key);
								
								Output output = null;
								
								if(p instanceof Double)
								{
									Double v = (Double)p;
									
									output = new Output();
									output.time = timestamp;
									output.power = v.intValue();
									output.useEnergy = true;
									output.energy_change = -1;
									output.energy = -1;
									
									outputs.add(output);
								}
								else if(p instanceof Long)
								{
									Long v = (Long)p;
									
									output = new Output();
									output.time = timestamp;
									output.power = v.intValue();
									output.useEnergy = true;
									output.energy_change = -1;
									output.energy = -1;
									
									outputs.add(output);
								}
								
								if(output != null)
								{
									String vkey = (String)properties.get("voltage");
									
									if(vkey != null && vkey.length() > 0)
									{
										Object ovolt = o.get(vkey);
										
										if(ovolt instanceof Double)
										{
											output.voltage = ((Double)ovolt).doubleValue();
										}
									}
									
									if(v7key != null)
										addExtendedData(output, v7key, jsonObject, 7);
									
									if(v8key != null)
										addExtendedData(output, v8key, jsonObject, 8);
									
									if(v9key != null)
										addExtendedData(output, v9key, jsonObject, 9);
									
									if(v10key != null)
										addExtendedData(output, v10key, jsonObject, 10);
									
									if(v11key != null)
										addExtendedData(output, v11key, jsonObject, 11);
									
									if(v12key != null)
										addExtendedData(output, v12key, jsonObject, 12);
									
									addStateOfCharge(output, (String)properties.get("soc-url"), (String)properties.get("soc-parameter"));
								}
							}
						}
					}
					else if(sensor.direction == IMPORT)
					{
						// consumption
						if(jsonObject.containsKey("load"))
						{
							JSONObject o = (JSONObject)jsonObject.get("load");
							Date timestamp = null;
							
							Object otime = o.get("last_communication_time");
							
							if(otime instanceof String)
							{
								String stime = (String)otime;
								
								if(stime.length() >= 19)
								{
									try
									{
										if(!stime.endsWith("Z"))
										{
											// non-utc timezone
											int toffset = stime.lastIndexOf('+');
											
											String tz = "";
											
											if(toffset > 0)
											{
												tz = stime.substring(toffset, stime.length());
											}
											else
											{
												toffset = stime.lastIndexOf('-');
												
												if(toffset > 0)
												{
													tz = stime.substring(toffset, stime.length());
												}
											}
											
											df.setTimeZone(TimeZone.getTimeZone("GMT" + tz));
										}
										
										timestamp = df.parse(stime.substring(0, 19));
									}
									catch(Exception e)
									{
										LOGGER.error("Invalid date format: " + stime.substring(0, 19));
									}
								}
							}
							
							if(timestamp != null)
							{
								String key = (String)properties.get("power");
								
								if(key == null)
								{
									key = "instant_power";
								}
								
								Object p = o.get(key);
								
								Output output = null;
								
								if(p instanceof Double)
								{
									Double v = (Double)p;
									
									output = new Output();
									output.time = timestamp;
									output.power = v.intValue();
									output.useEnergy = true;
									output.energy_change = -1;
									output.energy = -1;
									
									outputs.add(output);
								}
								else if(p instanceof Long)
								{
									Long v = (Long)p;
									
									output = new Output();
									output.time = timestamp;
									output.power = v.intValue();
									output.useEnergy = true;
									output.energy_change = -1;
									output.energy = -1;
									
									outputs.add(output);
								}
								
								if(output != null)
								{
									String vkey = (String)properties.get("voltage");
									
									if(vkey != null && vkey.length() > 0)
									{
										Object ovolt = o.get(vkey);
										
										if(ovolt instanceof Double)
										{
											output.voltage = ((Double)ovolt).doubleValue();
										}
									}
									
									if(v7key != null)
										addExtendedData(output, v7key, jsonObject, 7);
									
									if(v8key != null)
										addExtendedData(output, v8key, jsonObject, 8);
									
									if(v9key != null)
										addExtendedData(output, v9key, jsonObject, 9);
									
									if(v10key != null)
										addExtendedData(output, v10key, jsonObject, 10);
									
									if(v11key != null)
										addExtendedData(output, v11key, jsonObject, 11);
									
									if(v12key != null)
										addExtendedData(output, v12key, jsonObject, 12);
									
									addStateOfCharge(output, (String)properties.get("soc-url"), (String)properties.get("soc-parameter"));
								}
							}
						}
					}
				}
				else
				{
					LOGGER.info("Data is empty [" + s + "]");
				}
			}
			catch(Exception e)
			{
				LOGGER.error("Could not read data [" + s + "]", e);
			}

			return outputs;
		}

		private void addStateOfCharge(Output output, String url, String param) 
		{
			if(url != null && url.startsWith("http"))
			{
				LOGGER.info(">>> " + url);
				
				String s = WebClient.get(url);
				
				if(s != null)
				{
					try
					{
						JSONObject jsonObject = (JSONObject)JSONValue.parse(s);
						
						if(jsonObject != null)
						{
							Object o = jsonObject.get("percentage");
							
							Double v = Double.NaN;
							
							if(o instanceof Double)
							{
								v = (Double)o;
							}
							else if(o instanceof Long)
							{
								v = ((Long)o).doubleValue();
							}
							
							if(!v.isNaN())
							{
								if("v7".equals(param))
								{
									output.v7 = v;
								}
								else if("v8".equals(param))
								{
									output.v8 = v;
								}
								else if("v9".equals(param))
								{
									output.v9 = v;
								}
								else if("v10".equals(param))
								{
									output.v10 = v;
								}
								else if("v11".equals(param))
								{
									output.v11 = v;
								}
								else if("v12".equals(param))
								{
									output.v12 = v;
								}
								else
								{
									LOGGER.error("Unsupported [soc-parameter] value: " + param + ", expecting v7, v8, v9, v10, v11 or v12");
								}
							}
						}
					}
					catch(Exception e)
					{
						LOGGER.error("Could not soc read data [" + s + "]", e);
					}
				}
			}
		}

		private void addExtendedData(Output output, String key, JSONObject j, int param) 
		{
			String[] k = key.split("\\.");
			
			if(k != null && k.length == 2)
			{
				if(j.containsKey(k[0]))
				{
					JSONObject o = (JSONObject)j.get(k[0]);
					
					Object p = o.get(k[1]);
					
					Double v = Double.NaN;
					
					if(p instanceof Long)
					{
						v = ((Long)p).doubleValue();
						
						
					}
					else if(p instanceof Double)
					{
						v = (Double)p;
					}
					
					if(!v.isNaN())
					{
						if(param == 7)
						{
							output.v7 = v;
						}
						else if(param == 8)
						{
							output.v8 = v;
						}
						else if(param == 9)
						{
							output.v9 = v;
						}
						else if(param == 10)
						{
							output.v10 = v;
						}
						else if(param == 11)
						{
							output.v11 = v;
						}
						else if(param == 12)
						{
							output.v12 = v;
						}
					}
				}
			}
		}
	}
}
